from . import gw_eccentricity
from . import load_data
__author__ = gw_eccentricity.__author__
__email__ = gw_eccentricity.__email__
__copyright__ = gw_eccentricity.__copyright__
__license__ = gw_eccentricity.__license__
__version__ = gw_eccentricity.__version__
__doc__ = gw_eccentricity.__doc__

from .gw_eccentricity import measure_eccentricity
from .gw_eccentricity import get_available_methods
from .load_data import load_waveform
